import asyncio
import os
import sys

from prompt_toolkit import PromptSession
from prompt_toolkit.application import create_app_session_from_tty

from .provider import get_provider

ollama = get_provider("ollama")()


class ShitPluginException(Exception):
    ...


async def prompt_user(default_prompt: str | None) -> None:
    """
    Coroutine function to prompt the user for input, optionally preloading with a default value.

    :param default_prompt: The default text to preload into the prompt
    :return: None
    """
    with create_app_session_from_tty():
        prompt_session = PromptSession()
        prompt_session.output.write_raw("\033[E")  # Can I do this with one of the builtin methods
        query = await prompt_session.prompt_async("> ", default=default_prompt or "")
        async for item in ollama.generate(query):
            sys.stdout.write(item)


async def generate_response(query: str) -> None:
    """Generate a shell command based on the query.

    :param query: The query to generate a shell command for
    :return: None
    """
    try:
        ...
    except Exception as e:
        raise ShitPluginException("Uh oh! Shit plugin alert. The LLM plugin you are using shit the bed.") from e


def get_query_from_pipe() -> str | None:
    """Get a query from stdin pipe.

    :raises ValueError: If the input is more than one line
    :return: The query from the stdin pipe
    """
    if os.isatty(0):  # Return 'None' if nothing is in the pipe
        return None
    if not (lines := sys.stdin.readlines()):
        return None
    if len(lines) > 1:
        raise ValueError("Multi-line input not supported")
    return lines[0].rstrip()


def main() -> None:
    """
    This is the core function of the application.

    It accepts queries either through a standard input pipe or an interactive prompt.
    If there is a query from the input pipe, it processes the query to generate a response.
    If there isn't a query from the input pipe, it prompts the user for input.

    Environment variables:
        - SHOR_DEFAULT_PROMPT: This is the initial user prompt that can be configured via this environment variable.

    :returns: None
    """
    if query := get_query_from_pipe():
        asyncio.run(
            generate_response(query)
        )
        return

    default_prompt = os.environ.get("SHOR_DEFAULT_PROMPT")
    asyncio.run(
        prompt_user(default_prompt)
    )


if __name__ == '__main__':
    try:
        main()
    except (EOFError, KeyboardInterrupt):
        exit(0)



import tomllib
from collections.abc import Callable
from pathlib import Path
from threading import RLock
from typing import Any


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


_config_lock = RLock()


def _synchronized(func: Callable) -> Callable:
    """
    Synchronizes config file operations

    :param func: function to be synchronized
    :return: None
    """

    def wrapper(*args, **kwargs: Any) -> Any:
        _config_lock.acquire()
        try:
            return func(*args, **kwargs)
        finally:
            _config_lock.release()

    return wrapper


class Configuration:
    config_dir = Path.home() / "Library/Application Support" / "shelloracle"

    def __enter__(self):
        self.config_dir.mkdir(exist_ok=True)
        _config_lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _config_lock.release()


"""
Let's prototype some of the Configuration usage

Priority:
1. Command line arguments
2. Environment variables
3. Configuration file

# Do I like the config as a context manager? 
with Configuration(in_memory?=True) as config:
    a = config.a
    b = config.b
"""

if __name__ == '__main__':
    with open("../config.toml", "rb") as config_file:
        config = tomllib.load(config_file)
        print(config)



from abc import abstractmethod
from collections.abc import AsyncGenerator
from typing import Protocol


class Provider(Protocol):
    """
    LLM Provider Protocol

    All providers must implement this interface.
    """

    @abstractmethod
    def generate(self, prompt: str) -> AsyncGenerator[str, None, None]:
        """
        This is an asynchronous generator method which defines the protocol that a provider implementation
        should adhere to. The method takes a prompt as an argument and produces an asynchronous stream
        of string results.

        :param prompt: A string value which serves as input to the provider's process of generating results.
        :return: An asynchronous generator yielding string results.
        """


def get_provider(name: str) -> type[Provider]:
    from .ollama import Ollama
    providers = {
        "ollama": Ollama
    }
    return providers[name]



import json
from collections.abc import AsyncGenerator
from dataclasses import dataclass

import requests

from . import Provider

host = "localhost"
port = 11434
endpoint = f"http://{host}:{port}/api/generate"

model = "codellama:13b"
system_prompt = """Based on the following user description, generate a corresponding Bash command. Focus solely on 
interpreting the requirements and translating them into a single, executable Bash command. Ensure accuracy and 
relevance to the user's description. The output should be a valid Bash command that directly aligns with the user's 
intent, ready for execution in a command-line environment. Output nothing except for the command. No code block, no 
English explanation, no start/end tags."""


@dataclass
class GenerateRequest:
    model: str
    """(required) the model name"""
    prompt: str | None = None
    """the prompt to generate a response for"""
    format: str | None = None
    """the format to return a response in. Currently the only accepted value is json"""
    options: dict | None = None
    """additional model parameters listed in the documentation for the Modelfile such as temperature"""
    system: str | None = None
    """system prompt to (overrides what is defined in the Modelfile)"""
    template: str | None = None
    """the full prompt or prompt template (overrides what is defined in the Modelfile)"""
    context: str | None = None
    """the context parameter returned from a previous request to /generate, this can be used to keep a short 
    conversational memory"""
    stream: bool | None = None
    """if false the response will be returned as a single response object, rather than a stream of objects"""
    raw: bool | None = None
    """if true no formatting will be applied to the prompt and no context will be returned. You may choose to use 
    the raw parameter if you are specifying a full templated prompt in your request to the API, and are managing 
    history yourself. JSON mode"""

    def to_json(self):
        request = {key: value for key, value in self.__dict__.items() if value is not None}
        return json.dumps(request)


class Ollama(Provider):

    async def generate(self, prompt: str) -> AsyncGenerator[str, None, None]:
        request = GenerateRequest(model, prompt, system=system_prompt, stream=True)
        r = requests.post(endpoint, data=request.to_json(), stream=True)

        if r.encoding is None:
            r.encoding = 'utf-8'

        for line in r.iter_lines(decode_unicode=True):
            if line:
                yield json.loads(line)["response"]


if __name__ == '__main__':
    ollama = Ollama()
